@@ -61,7 +61,7 @@ class CompositeHelper(AutoPropGObjectMixin, gobject.GObject):
         remove_event_receiver(self._window, self)
         self._window = None
 
-    def acknowledge_changes(self, x, y, w, h):
+    def acknowledge_changes(self):
         if self._damage_handle is not None:
             trap.swallow(xdamage_acknowledge,
                          self._window, self._damage_handle)
@@ -75,9 +75,10 @@ class CompositeHelper(AutoPropGObjectMixin, gobject.GObject):
 
     def _cleanup_listening(self, listening):
         # Don't want to stop listening to self._window!:
-        assert self._window not in self._listening_to
-        for w in self._listening_to:
-            remove_event_receiver(w, self)
+        if self._listening_to:
+            assert self._window not in self._listening_to
+            for w in self._listening_to:
+                remove_event_receiver(w, self)
 
     def do_get_property_contents_handle(self, name):
         if self._contents_handle is None:
