@@ -9,6 +9,7 @@ import gtk
 import gobject
 import cairo
 import re
+import os
 
 from wimpiggy.util import (n_arg_signal,
                            gtk_main_quit_really,
@@ -32,11 +33,16 @@ def nn(x):
 
 class ClientSource(object):
     def __init__(self, protocol):
+        self._priority_packets = []
         self._ordinary_packets = []
         self._mouse_position = None
         self._protocol = protocol
         self._protocol.source = self
 
+    def queue_priority_packet(self, packet):
+        self._priority_packets.append(packet)
+        self._protocol.source_has_more()
+
     def queue_ordinary_packet(self, packet):
         self._ordinary_packets.append(packet)
         self._protocol.source_has_more()
@@ -50,16 +56,19 @@ class ClientSource(object):
         self._protocol.source_has_more()
 
     def next_packet(self):
-        if self._ordinary_packets:
+        if self._priority_packets:
+            packet = self._priority_packets.pop(0)
+        elif self._ordinary_packets:
             packet = self._ordinary_packets.pop(0)
-            return packet, (bool(self._ordinary_packets)
-                            or self._mouse_position is not None)
         elif self._mouse_position is not None:
             packet = self._mouse_position
             self._mouse_position = None
-            return packet, False
         else:
-            return None, False
+            packet = None
+        has_more = packet is not None and \
+                (bool(self._priority_packets) or bool(self._ordinary_packets) \
+                 or self._mouse_position is not None)
+        return packet, has_more
 
 class ClientWindow(gtk.Window):
     def __init__(self, client, id, x, y, w, h, metadata, override_redirect):
@@ -188,7 +197,21 @@ class ClientWindow(gtk.Window):
 
     def draw(self, x, y, width, height, coding, img_data):
         gc = self._backing.new_gc()
-        if coding != "rgb24":
+        if coding == "mmap":
+            assert self._client.supports_mmap
+            log("drawing from mmap: %s", img_data)
+            data = ""
+            import ctypes
+            data_start = ctypes.c_uint.from_buffer(self._client.mmap, 0)
+            for offset, length in img_data:
+                self._client.mmap.seek(offset)
+                data += self._client.mmap.read(length)
+                data_start.value = offset+length
+            self._backing.draw_rgb_image(gc, x, y, width, height, gtk.gdk.RGB_DITHER_NONE, data)
+        elif coding == "rgb24":
+            assert len(img_data) == width * height * 3
+            self._backing.draw_rgb_image(gc, x, y, width, height, gtk.gdk.RGB_DITHER_NONE, img_data)
+        else:
             loader = gtk.gdk.PixbufLoader(coding)
             loader.write(img_data, len(img_data))
             loader.close()
@@ -201,9 +224,6 @@ class ClientWindow(gtk.Window):
                     self._failed_pixbuf_index += 1
             else:
                 self._backing.draw_pixbuf(gc, pixbuf, 0, 0, x, y, width, height)
-        else:
-            assert len(img_data) == width * height * 3
-            self._backing.draw_rgb_image(gc, x, y, width, height, gtk.gdk.RGB_DITHER_NONE, img_data)
         self.window.invalidate_rect(gtk.gdk.Rectangle(x, y, width, height), False)
 
         if self._refresh_requested:
@@ -349,6 +369,37 @@ class XpraClient(gobject.GObject):
         self.notifications_enabled = True
         self._client_extras = ClientExtras(self, opts)
         self.clipboard_enabled = opts.clipboard and self._client_extras.supports_clipboard()
+        self.send_damage_sequence = False
+        self.mmap_enabled = False
+        self.supports_mmap = opts.mmap and self._client_extras.supports_mmap()
+        self.mmap = None
+        self.mmap_file = None
+        self.mmap_size = 0
+        if self.supports_mmap:
+            try:
+                import mmap
+                import tempfile
+                dotxpradir = os.path.expanduser("~/.xpra")
+                if not os.path.exists(dotxpradir):
+                    os.mkdir(dotxpradir, 0700)
+                temp = tempfile.NamedTemporaryFile(prefix="xpra.", suffix=".mmap", dir=dotxpradir)
+                #keep a reference to it so it does not disappear!
+                self._mmap_temp_file = temp
+                self.mmap_file = temp.name
+                self.mmap_size = max(4096, mmap.PAGESIZE)*32*1024   #generally 128MB
+                fd = temp.file.fileno()
+                log("using mmap file %s, fd=%s, size=%s", self.mmap_file, fd, self.mmap_size)
+                os.lseek(fd, self.mmap_size-1, os.SEEK_SET)
+                assert os.write(fd, '\x00')
+                os.lseek(fd, 0, os.SEEK_SET)
+                self.mmap = mmap.mmap(fd, length=self.mmap_size)
+            except Exception, e:
+                log.error("failed to setup mmap: %s", e)
+                self.supports_mmap = False
+                self.clean_mmap()
+                self.mmap = None
+                self.mmap_file = None
+                self.mmap_size = 0
 
         self._protocol = Protocol(conn, self.process_packet)
         ClientSource(self._protocol)
@@ -395,11 +446,18 @@ class XpraClient(gobject.GObject):
     def run(self):
         gtk_main_quit_on_fatal_exceptions_enable()
         gtk.main()
+
+    def cleanup(self):
         self._client_extras.exit()
         if self._protocol:
             self._protocol.close()
+        self.clean_mmap()
 
-    def quit(self):
+    def clean_mmap(self):
+        if self.mmap_file and os.path.exists(self.mmap_file):
+            os.unlink(self.mmap_file)
+
+    def quit(self, *args):
         gtk_main_quit_really()
 
     def parse_shortcuts(self, strs):
@@ -504,7 +562,7 @@ class XpraClient(gobject.GObject):
                 #confirm it and continue, otherwise stop
                 log.debug("continue_key_repeat")
                 if keycode in self.keycodes_pressed:
-                    self.send(["key-repeat", keycode])
+                    self.send_now(["key-repeat", keycode])
                     return  True
                 else:
                     del self.keycodes_pressed[keycode]
@@ -514,12 +572,12 @@ class XpraClient(gobject.GObject):
                 #confirm it and start repeat:
                 log.debug("start_key_repeat")
                 if keycode in self.keycodes_pressed:
-                    self.send(["key-repeat", keycode])
+                    self.send_now(["key-repeat", keycode])
                     self.keycodes_pressed[keycode] = gobject.timeout_add(interval, continue_key_repeat)
                 else:
                     del self.keycodes_pressed[keycode]
                 return  False   #never run this timer again
-            log.debug("key repeat: starting timer for %s with delay %s and interval %s", keycode, delay, start_key_repeat)
+            log.debug("key repeat: starting timer for %s with delay %s and interval %s", keycode, delay, interval)
             self.keycodes_pressed[keycode] = gobject.timeout_add(delay, start_key_repeat)
 
     def clear_repeat(self):
@@ -580,6 +638,9 @@ class XpraClient(gobject.GObject):
     def send(self, packet):
         self._protocol.source.queue_ordinary_packet(packet)
 
+    def send_now(self, packet):
+        self._protocol.source.queue_priority_packet(packet)
+
     def send_positional(self, packet):
         self._protocol.source.queue_positional_packet(packet)
 
@@ -620,9 +681,12 @@ class XpraClient(gobject.GObject):
         root_w, root_h = gtk.gdk.get_default_root_window().get_size()
         capabilities_request["desktop_size"] = [root_w, root_h]
         capabilities_request["png_window_icons"] = True
+        capabilities_request["damage_sequence"] = True
         key_repeat = self._client_extras.get_keyboard_repeat()
         if key_repeat:
             capabilities_request["key_repeat"] = key_repeat
+        if self.mmap_file:
+            capabilities_request["mmap_file"] = self.mmap_file
         self.send(["hello", capabilities_request])
 
     def send_jpeg_quality(self, q):
@@ -687,8 +751,8 @@ class XpraClient(gobject.GObject):
                 log.warn("Server's virtual screen is too small -- "
                          "(server: %sx%s vs. client: %sx%s)\n"
                          "You may see strange behavior.\n"
-                         "Please complain to "
-                         "parti-discuss@partiwm.org"
+                         "Please see "
+                         "http://xpra.org/trac/ticket/10"
                          % (avail_w, avail_h, root_w, root_h))
         self._protocol._send_size = capabilities.get("packet_size", False)
         randr = capabilities.get("resize_screen", False)
@@ -708,6 +772,12 @@ class XpraClient(gobject.GObject):
         self.notifications_enabled = capabilities.get("notifications", False)
         clipboard_server_support = capabilities.get("clipboard", True) 
         self.clipboard_enabled = clipboard_server_support and self._client_extras.supports_clipboard()
+        self.send_damage_sequence = capabilities.get("damage_sequence", False)
+        self.mmap_enabled = self.supports_mmap and self.mmap_file and capabilities.get("mmap_enabled")
+        if self.mmap_enabled:
+            log.info("mmap enabled using %s", self.mmap_file)
+        #the server will have a handle on the mmap file by now, safe to delete:
+        self.clean_mmap()
         #ui may want to know this is now set:
         self.emit("clipboard-toggled")
         self.key_repeat_delay, self.key_repeat_interval = capabilities.get("key_repeat", (-1,-1))
@@ -745,9 +815,15 @@ class XpraClient(gobject.GObject):
         self._process_new_common(packet, True)
 
     def _process_draw(self, packet):
-        (_, id, x, y, width, height, coding, data) = packet
+        (id, x, y, width, height, coding, data) = packet[1:8]
+        if len(packet)==9:
+            packet_sequence = packet[8]
+        else:
+            packet_sequence = None
         window = self._id_to_window[id]
         window.draw(x, y, width, height, coding, data)
+        if packet_sequence and self.send_damage_sequence:
+            self.send_now(["damage-sequence", packet_sequence])
 
     def _process_cursor(self, packet):
         (_, new_cursor) = packet
