@@ -11,9 +11,6 @@
 XPRA_LOCAL_SERVERS_SUPPORTED = False
 DEFAULT_SSH_CMD = "ssh"
 GOT_PASSWORD_PROMPT_SUGGESTION = "Perhaps you need to set up your ssh agent?\n"
-X11_KEYMAPS = False
 
 def add_client_options(parser):
-    parser.add_option("--dock-icon", action="store",
-                          dest="dock_icon", default=None,
-                          help="Path to the image which will be used as icon for the dock")
+    pass
