@@ -14,7 +14,7 @@ import os
 import socket # for socket.error
 import zlib
 
-from Queue import Queue, Empty
+from Queue import Queue
 from threading import Thread, Lock
 
 from xpra.bencode import bencode, IncrBDecode
@@ -60,7 +60,7 @@ def repr_ellipsized(obj, limit=100):
         return repr(obj)
 
 def dump_packet(packet):
-    return "[" + ", ".join([repr_ellipsized(x, 50) for x in packet]) + "]"
+    return "[" + ", ".join([repr_ellipsized(str(x), 50) for x in packet]) + "]"
 
 def main_thread_call(fn, *args, **kwargs):
     log("Queueing main thread call to %s" % (fn,))
@@ -101,7 +101,7 @@ class Protocol(object):
         self._conn = conn
         self._process_packet_cb = process_packet_cb
         self._write_queue = Queue()
-        self._read_queue = Queue()
+        self._read_queue = Queue(5)
         # Invariant: if .source is None, then _source_has_more == False
         self.source = None
         self.jpegquality = 0
@@ -112,13 +112,16 @@ class Protocol(object):
         self._read_decoder = IncrBDecode()
         self._compressor = None
         self._decompressor = None
-        self._read_loop_count = CachedCounter()
         self._write_thread = Thread(target=self._write_thread_loop)
         self._write_thread.daemon = True
         self._write_thread.start()
+        self._write_lock = Lock()
         self._read_thread = Thread(target=self._read_thread_loop)
         self._read_thread.daemon = True
         self._read_thread.start()
+        self._read_parser_thread = Thread(target=self._read_parse_thread_loop)
+        self._read_parser_thread.deamon = True
+        self._read_parser_thread.start()
         self._maybe_queue_more_writes()
 
     def source_has_more(self):
@@ -150,16 +153,20 @@ class Protocol(object):
             return
         packet, self._source_has_more = self.source.next_packet()
         if packet is not None:
-            log("writing %s", dump_packet(packet), type="raw.write")
+            #log("writing %s", dump_packet(packet), type="raw.write")
             data = bencode(packet)
             l = len(data)
-            if l<=1024:
-                #send size and data together (low copy overhead):
-                self._queue_write("PS%014d%s" % (l, data), True)
-            else:
+            self._write_lock.acquire()
+            try:
                 if self._send_size:
+                    if l<=1024:
+                        #send size and data together (low copy overhead):
+                        self._queue_write("PS%014d%s" % (l, data), True)
+                        return
                     self._queue_write("PS%014d" % l)
                 self._queue_write(data, True)
+            finally:
+                self._write_lock.release()
 
     def _write_thread_loop(self):
         try:
@@ -172,7 +179,6 @@ class Protocol(object):
                     break
                 try:
                     while buf:
-                        log("write thread: writing %s", repr_ellipsized(buf))
                         buf = buf[self._conn.write(buf):]
                 except (OSError, IOError, socket.error), e:
                     main_thread_call(self._connection_lost, "Error writing to connection: %s" % e)
@@ -187,31 +193,25 @@ class Protocol(object):
             self._conn.close()
 
     def _read_thread_loop(self):
-        last_scheduled_read = -1
-        while not self._closed:
-            log("read thread: waiting for data to arrive")
-            try:
-                buf = self._conn.read(8192)
-            except (ValueError, OSError, IOError, socket.error), e:
-                main_thread_call(self._connection_lost, "Error reading from connection: %s" % e)
-                return
-            except TypeError:
-                assert self._closed
-                return
-            log("read thread: got data %s", repr_ellipsized(buf))
-            self._recv_counter += len(buf)
-            self._read_queue.put(buf)
-            #schedule the read thread to run if it has not fired since we started
-            #reading in this loop (-1!=0), or if it has already fired since
-            #the last time we scheduled it to run:
-            c = self._read_loop_count.value()
-            if last_scheduled_read!=c:
-                last_scheduled_read = c
-                main_thread_call(self._handle_read)
-            if not buf:
-                log("read thread: eof")
-                break
-        log("read thread: ended")
+        try:
+            while not self._closed:
+                log("read thread: waiting for data to arrive")
+                try:
+                    buf = self._conn.read(8192)
+                except (ValueError, OSError, IOError, socket.error), e:
+                    main_thread_call(self._connection_lost, "Error reading from connection: %s" % e)
+                    return
+                except TypeError:
+                    assert self._closed
+                    return
+                log("read thread: got data %s", repr_ellipsized(buf))
+                self._recv_counter += len(buf)
+                self._read_queue.put(buf)
+                if not buf:
+                    log("read thread: eof")
+                    break
+        finally:
+            log("read thread: ended")
 
     def _connection_lost(self, message="", exc_info=False):
         log.info("connection lost: %s", message, exc_info=exc_info)
@@ -219,48 +219,38 @@ class Protocol(object):
             self._process_packet_cb(self, [Protocol.CONNECTION_LOST])
             self.close()
 
-    def _handle_read(self):
-        if self._closed:
-            return
-        log("main thread: woken to handle read data")
-        while not self._closed:
-            self._read_loop_count.inc()
-            try:
-                buf = self._read_queue.get(block=False)
-                log("main thread: found read data %s", repr_ellipsized(buf))
-            except Empty:
-                log("main thread: processed all read data")
-                return
-            if not buf:
-                self._connection_lost("empty marker in read queue")
-                return
-            if self._decompressor is not None:
-                buf = self._decompressor.decompress(buf)
-            try:
-                self._read_decoder.add(buf)
-            except:
-                self._connection_lost("read buffer is in an inconsistent state, cannot continue", exc_info=True)
-                return
+    def _read_parse_thread_loop(self):
+        try:
             while not self._closed:
-                had_deflate = (self._decompressor is not None)
+                buf = self._read_queue.get()
+                if not buf:
+                    return self._connection_lost("empty marker in read queue")
+                if self._decompressor is not None:
+                    buf = self._decompressor.decompress(buf)
                 try:
-                    result = self._read_decoder.process()
-                except ValueError:
-                    # Peek at the data we got, in case we can make sense of it:
-                    self._process_packet([Protocol.GIBBERISH, self._read_decoder.unprocessed()])
-                    # Then hang up:
-                    self._connection_lost("gibberish received")
-                    return
-                if result is None:
-                    break
-                packet, unprocessed = result
-                self._process_packet(packet)
-                if not had_deflate and (self._decompressor is not None):
-                    # deflate was just enabled: so decompress the unprocessed
-                    # data
-                    unprocessed = self._decompressor.decompress(unprocessed)
-                self._read_decoder = IncrBDecode(unprocessed)
-        log("main thread: read handled")
+                    self._read_decoder.add(buf)
+                except:
+                    return self._connection_lost("read buffer is in an inconsistent state, cannot continue", exc_info=True)
+                while not self._closed:
+                    had_deflate = (self._decompressor is not None)
+                    try:
+                        result = self._read_decoder.process()
+                    except ValueError:
+                        # Peek at the data we got, in case we can make sense of it:
+                        self._process_packet([Protocol.GIBBERISH, self._read_decoder.unprocessed()])
+                        # Then hang up:
+                        return self._connection_lost("gibberish received")
+                    if result is None:
+                        break
+                    packet, unprocessed = result
+                    main_thread_call(self._process_packet, packet)
+                    if not had_deflate and (self._decompressor is not None):
+                        # deflate was just enabled: so decompress the unprocessed
+                        # data
+                        unprocessed = self._decompressor.decompress(unprocessed)
+                    self._read_decoder = IncrBDecode(unprocessed)
+        finally:
+            log("read parse thread: ended")
 
     def _process_packet(self, decoded):
         if self._closed:
@@ -288,5 +278,10 @@ class Protocol(object):
 
     def close(self):
         if not self._closed:
-            self._write_queue.put(None)
             self._closed = True
+        #make the threads exit by adding the empty marker:
+        self._write_queue.put(None)
+        try:
+            self._read_queue.put_nowait(None)
+        except:
+            pass
