@@ -138,6 +138,9 @@ def main(script_file, cmdline):
                       dest="key_shortcuts", type="str", default=[],
                       help="Define key shortcuts that will trigger specific actions."
                       + " Defaults to Meta+Shift+F4:quit if no shortcuts are defined.")
+    parser.add_option("--disable-mmap", action="store_false",
+                      dest="mmap", default=True,
+                      help="Disable memory mapped transfers for local connections")
     parser.add_option("-z", "--compress", action="store",
                       dest="compression_level", type="int", default=3,
                       metavar="LEVEL",
@@ -187,11 +190,8 @@ def main(script_file, cmdline):
         nox()
         from xpra.scripts.server import run_server
         run_server(parser, options, mode, script_file, args)
-    elif mode == "attach":
-        try:
-            run_client(parser, options, args)
-        except KeyboardInterrupt:
-            sys.stdout.write("Exiting on keyboard interrupt\n")
+    elif mode == "attach" or mode == "detach":
+        run_client(parser, options, args, mode=="detach")
     elif mode == "stop" and XPRA_LOCAL_SERVERS_SUPPORTED:
         nox()
         run_stop(parser, options, args)
@@ -297,7 +297,7 @@ def connect_or_fail(display_desc):
         assert False, "unsupported display type in connect"
 
 
-def run_client(parser, opts, extra_args):
+def run_client(parser, opts, extra_args, detach):
     from xpra.client import XpraClient
     conn = connect_or_fail(pick_display(parser, opts, extra_args))
     if opts.compression_level < 0 or opts.compression_level > 9:
@@ -322,7 +322,17 @@ def run_client(parser, opts, extra_args):
             sys.stdout.flush()
     app.connect("handshake-complete", handshake_complete_msg)
     app.connect("received-gibberish", got_gibberish_msg)
-    app.run()
+    if detach:
+        def do_detach(*args):
+            sys.stdout.write("handshake-complete: detaching")
+            app.quit()
+        app.connect("handshake-complete", do_detach)
+    import signal
+    signal.signal(signal.SIGINT, app.quit)
+    try:
+        app.run()
+    finally:
+        app.cleanup()
 
 def run_proxy(parser, opts, extra_args):
     from xpra.proxy import XpraProxy
