@@ -8,7 +8,7 @@
 XPRA_LOCAL_SERVERS_SUPPORTED = True
 DEFAULT_SSH_CMD = "ssh"
 GOT_PASSWORD_PROMPT_SUGGESTION = "Perhaps you need to set up your ssh agent?\n"
-X11_KEYMAPS = True
 
 def add_client_options(parser):
-    pass
+    from xpra.platform import add_notray_option
+    add_notray_option(parser)
